import numpy as np

from audiodag.signal.digital.conversion import ms_to_pts
from audiodag.signal.envelopes.envelope import Envelope

from dataclasses import dataclass


class ConstantEnvelope(Envelope):
    def f(self, y: np.ndarray) -> np.ndarray:
        return y


class CosEnvelope(Envelope):
    def f(self, y: np.ndarray) -> np.ndarray:
        return y * (np.cos(np.linspace(1 * np.pi, 3 * np.pi, len(y))) + 1) * 0.5


@dataclass
class CosRiseEnvelope(Envelope):
    rise: int
    fs: int

    @property
    def rise_pts(self) -> int:
        return ms_to_pts(fs=self.fs,
                         t_ms=self.rise)

    def f(self, y: np.ndarray) -> np.ndarray:
        cos_rise = (np.cos(np.linspace(1 * np.pi, 2 * np.pi, self.rise_pts)) + 1) * 0.5
        cos_fall = np.flip(cos_rise)

        uniform_centre = np.ones(shape=(len(y) - 2 * self.rise_pts))

        envelope = np.concatenate((cos_rise, uniform_centre, cos_fall),
                                  axis=0)

        return y * envelope


class IncreasingEnvelope(Envelope):
    def f(self, y):
        return y * np.linspace(0, 1, len(y))



import gc
from abc import ABC, abstractmethod
from typing import Union

import numpy as np
from matplotlib import pyplot as plt

from audiodag.common.repr_id import ReprID
from audiodag.signal.digital.conversion import ms_to_pts
from audiodag.signal.envelopes.templates import Envelope, ConstantEnvelope


class DigitalSignal(ReprID):
    """
    Object representing basic properties of a digital signal, handles time <-> samples conversion. Doesn't bother
    with with analog <-> bits scale on y yet, though.
    """

    def __init__(self,
                 start: int = 0,
                 fs: int = 1000,
                 duration: int = 20,
                 mag: int = 1,
                 clip: float = 2,
                 envelope: Envelope = ConstantEnvelope,
                 seed: Union[int, None] = None,
                 cache: bool = False) -> None:
        """

        :param start: Start time, in ms.
        :param fs: Sampling rate.
        :param duration: Duration in ms.
        :param mag: Magnitude of signal in volts.
        :param clip: Max positive magnitude of signal.
        :param seed: Integer used to set numpy RandomState used for generating stochastic signals.
        :param cache: If True, hold signal in memory after generation. Otherwise generate each time it's accessed.
        """
        self.start = start
        self.duration = duration
        self.mag = mag
        self.fs = fs
        self.seed = seed
        self.state = seed
        self.envelope = envelope
        self.cache = cache
        self.clip = clip

        self.envelope = envelope(fs=fs)

        self._y: Union[np.ndarray, None] = None
        self._seed: int
        self._state: np.random.RandomState

    def __repr__(self) -> str:
        """__repr__ is used for id and eq, it should be redefined in children."""
        return f"DigitalTime(fs={self.fs}, duration={self.duration}, seed={self.state})"

    def clear(self) -> None:
        """Remove the signal vector from memory."""
        self._y = None
        gc.collect()

    @abstractmethod
    def _generate_f(self) -> np.ndarray:
        """Function to generate Signal. This should be overloaded in child."""
        return np.zeros(shape=(self.duration_pts,))

    def _generate(self) -> np.ndarray:
        """
        Generate the signal.

        Combines the generator function, the envelope, and applies clipping.
        """
        y = self._generate_f()
        y = self.envelope(y)
        y[y > self.clip] = self.clip

        return y

    @property
    def x(self) -> np.ndarray:
        """Time axis, in ms."""
        return np.linspace(self.start, self.start + self.duration, self.duration_pts)

    @property
    def x_pts(self) -> np.ndarray:
        """Time axis, in samples."""
        return np.linspace(self.start_pts, self.start_pts + self.duration_pts - 1, self.duration_pts,
                           dtype=int)

    @property
    def y(self) -> np.ndarray:
        """Signal. Generated on the fly if required."""
        if self._y is None:
            self.state = self.seed
            y = self._generate()
        else:
            y = self._y

        if self.cache:
            self._y = y

        return y

    def plot(self,
             show: bool = False,
             **kwargs):
        """Plot the signal against time in ms."""
        plt.plot(self.x, self.y,
                 **kwargs)

        if show:
            plt.show()

    @property
    def duration_pts(self) -> int:
        """Duration of the signal, in ms."""
        return ms_to_pts(t_ms=self.duration,
                         fs=self.fs)

    @property
    def start_pts(self) -> float:
        """Duration of the signal, in samples."""
        return ms_to_pts(t_ms=self.start,
                         fs=self.fs)

    @property
    def seed(self) -> int:
        """Return the seed used to generate the signal."""
        return self._seed

    @seed.setter
    def seed(self, seed: Union[int, None]) -> None:
        """Set the seed by generating a RandomState from the input."""

        if seed is None:
            seed = np.random.RandomState(seed=seed).randint(2 ** 31)

        self.state = seed
        self._seed = seed

    @property
    def state(self) -> np.random.RandomState:
        return self._state

    @state.setter
    def state(self, seed) -> None:
        self._state = np.random.RandomState(seed)



from dataclasses import dataclass
from typing import Any, List, Tuple


@dataclass
class Edge:
    weight: float = 1.0


@dataclass
class Node:
    data: Any
    connections_out: List[Tuple[float, "Node"]]
    terminal: bool = False




import copy
from functools import reduce
from typing import List, Tuple, Callable, Iterable

import numpy as np
from matplotlib import pyplot as plt

from audiodag.signal.digital.conversion import pts_to_ms
from audiodag.signal.digital.digital_siginal import DigitalSignal
from audiodag.signal.envelopes.envelope import Envelope
from audiodag.signal.envelopes.templates import ConstantEnvelope


class Component(DigitalSignal):
    def __init__(self, weight: float = 1.0,
                 *args, **kwargs):
        """

        :param weight: Relative event weight, used when combined with other components, for example.
        """

        self.weight = weight
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Event(fs={self.fs}, start={self.start}, duration={self.duration}, weight={self.weight}, seed={self.seed})"

    def _generate_f(self) -> np.ndarray:
        """Default components is constant 1s * mag"""

        return np.ones(shape=(self.duration_pts,)) * self.mag

    def __mul__(self, other):
        """Multiplying components generates a CompoundEvent object with a generator for the combined signals."""

        return CompoundComponent(events=[self, other])


class CompoundComponent(Component):
    """
    Object for combining components, for example adding noise to another components.

    Supports combination of multiple components, but only with equal weighting and same durations for now.
    """

    def __init__(self, events: List[Component],
                 weights: List[float] = None,
                 start: int = None,
                 envelope: Envelope = ConstantEnvelope):

        self._verify_event_list(events)
        start_sub, _, duration = self._new_duration(events)

        super().__init__(fs=events[0].fs,
                         start=start_sub,
                         duration=pts_to_ms(duration,
                                            fs=events[0].fs),
                         envelope=envelope)

        self.events = []

        self._assign_events(events)
        self._assign_weights()

        # Adjust start of self and all sub components, if a new start is specified.
        if start is not None:
            self._adjust_start(start - start_sub)
            self.start = start

        self._generate_f = self._make_generate_f()

    def __repr__(self) -> str:
        return f"CombinedEvent(components={self.events})"

    def __hash__(self) -> int:
        return hash(str(self.events) + str(self.fs) + str(self.envelope.__name__)
                    + str(self.duration))

    def _adjust_start(self, start_delta: int):
        """
        Given a delta, recursively adjust start of all parent components.
        TODO: Recursiveness not tested....
        """

        for ev in self.events:
            if isinstance(ev, CompoundComponent):
                ev._adjust_start(start_delta)
            else:
                ev.start = ev.start + start_delta

    def _assign_events(self, events: List[Component]) -> None:
        # Events need to be copied here as weights may need to be updated
        # Shouldn't need to be a deep copy, though. Can also clear any arrays, may want to parameterize this.
        for ev in events:
            ev = copy.copy(ev)
            ev.clear()
            self.events.append(ev)

    @staticmethod
    def _normalise_to_sum_1(y: np.ndarray) -> np.ndarray:
        """Normalise vector y to a sum to 1."""
        return y / y.sum()

    def _assign_weights(self,
                        weights: List[float] = None) -> None:
        """
        For the provided components, reset their normalised, relative weights. Or override them with supplied.
        """
        if weights is None:
            # Set using values in components
            weights = [ev.weight for ev in self.events]

        # Normalise
        weights = self._normalise_to_sum_1(np.array(weights))

        # Update event weights with normalised
        for ev, w in zip(self.events, weights):
            ev.weight = w

    @staticmethod
    def _verify_event_list(events: List[Component]):
        check_params = ['fs']

        for p in check_params:
            param_values = [getattr(e, p) for e in events]
            if len(np.unique(param_values)) > 1:
                raise ValueError(f"Param {p} is inconsistent across components: {param_values}")

    @staticmethod
    def _new_duration(events: List[Component]):
        start = reduce(lambda ev_a, ev_b: min(ev_a, ev_b), [e.x_pts.min() for e in events])
        end = reduce(lambda ev_a, ev_b: max(ev_a, ev_b), [e.x_pts.max() for e in events])
        return start, end, end - start + 1

    def channels(self) -> np.ndarray:

        y = np.zeros(shape=(len(self.events), self.duration_pts))
        for e_i, e in enumerate(self.events):
            y[e_i, e.x_pts - self.start] = e.y * e.weight

        return y

    def _combiner(self) -> Tuple[np.ndarray, np.ndarray]:

        y = self.channels()
        y = y.sum(axis=0)

        return y

    def _make_generate_f(self) -> Callable:
        """
        Make the generator function.

        Returns as Callable that does the combination using the .y properties on each Event. This callable can be
        assigned to ._generate_f which maintains the Event API.
        """
        return self._combiner

    def plot(self,
             channels: bool = False,
             *args, **kwargs):

        if channels:
            plt.plot(self.x, self.channels().T)

        super().plot(*args, **kwargs)

    def plot_subplots(self,
                      show: bool = False):
        fig, ax = plt.subplots(nrows=len(self.events) + 1,
                               ncols=1)

        for e_i, e in enumerate(self.events):
            ax[e_i].plot(e.x, e.y,
                         color='k' if isinstance(e, CompoundComponent) else 'r')
            ax[e_i].set_xlim([self.x[0], self.x[-1]])
            ax[e_i].set_title(f"Signal component {e_i}")

        ax[-1].plot(self.x, self.y)
        ax[-1].set_title('Combined signal')

        if show:
            plt.show()

    def to_list(self) -> List[Tuple[int, Component]]:
        return self.recursive_transverse(self)

    def recursive_transverse(self, ev: Component,
                             depth: int = 0,
                             path: int = 0,
                             previous_node: str='|') -> Iterable[Tuple[int, Component]]:
        """Currently just prints a very crude representation of the graph to the console."""

        node = f"{previous_node} <- {path}({str(depth)})"

        if isinstance(ev, CompoundComponent):
            return [self.recursive_transverse(e,
                                              depth=depth + 1,
                                              path=p,
                                              previous_node=node) for p, e in enumerate(ev.events)]
        else:
            return node, ev



import unittest

import numpy as np

from audiodag.signal.envelopes.templates import ConstantEnvelope, CosEnvelope, CosRiseEnvelope


ONES = np.ones(shape=(1000, ))
ZEROS = np.zeros(shape=(800, ))


class TestConstantEnvelope(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.env = ConstantEnvelope()

    def test_envelope_on_ones(self) -> None:
        self.assertTrue(np.all(self.env(ONES) == ONES))

    def test_envelope_on_zeros(self) -> None:
        self.assertTrue(np.all(self.env(ZEROS) == ZEROS))


class TestCosEnvelope(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.env = CosEnvelope()

    def test_envelope_on_ones(self) -> None:
        self.assertLess(np.sum(self.env(ONES)), np.sum(ONES))

    def test_envelope_on_zeros(self) -> None:
        self.assertTrue(np.all(np.round(self.env(ZEROS)) == ZEROS))


class TestCosRiseEnvelope(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.env = CosRiseEnvelope(fs=1000,
                                  rise=200)

    def test_envelope_on_ones(self) -> None:
        self.assertLess(np.sum(self.env(ONES)), np.sum(ONES))

    def test_envelope_on_zeros(self) -> None:
        self.assertTrue(np.all(self.env(ZEROS) == ZEROS))


import unittest
from unittest.mock import patch

import numpy as np
from matplotlib import pyplot as plt

from audiodag.signal.digital.digital_siginal import DigitalSignal
from audiodag.signal.digital.conversion import pts_to_ms, ms_to_pts


class TestFunctions(unittest.TestCase):
    def test_ms_to_pts(self):
        self.assertEqual(ms_to_pts(t_ms=1000,
                                   fs=1000), 1000)

        self.assertEqual(ms_to_pts(t_ms=500,
                                   fs=1000), 500)

        self.assertEqual(ms_to_pts(t_ms=1000,
                                   fs=500), 500)

        self.assertEqual(ms_to_pts(t_ms=20,
                                   fs=800), 16)

    def test_pts_to_ms(self):
        self.assertEqual(pts_to_ms(t_pts=1000,
                                   fs=1000), 1000)

        self.assertEqual(pts_to_ms(t_pts=500,
                                   fs=1000), 500)


class TestDigitalSignal(unittest.TestCase):

    @classmethod
    @patch.multiple(DigitalSignal, __abstractmethods__=set())
    def setUpClass(cls) -> None:
        cls.dt_1 = DigitalSignal(fs=100,
                                 duration=1000)
        cls.dt_2 = DigitalSignal(fs=100,
                                 duration=500)

    def test_eq(self):
        self.assertTrue(self.dt_1 == self.dt_1)
        self.assertFalse(self.dt_1 == self.dt_2)

    def test_unique(self):
        self.assertEqual(len(np.unique([self.dt_1, self.dt_1])), 1)
        self.assertEqual(len(np.unique([self.dt_1, self.dt_2])), 2)

    def test_duration_in_pts(self):
        self.assertEqual(self.dt_1.duration_pts, 100)

    @patch.multiple(DigitalSignal, __abstractmethods__=set())
    def test_with_defaults(self):
        dt = DigitalSignal()

        self.assertEqual(len(dt.x), 20)
        self.assertEqual(len(dt.x_pts), 20)

    def test_plot(self):
        self.dt_1.plot()
        plt.show()


from audiodag.signal.digital.conversion import db_to_lin
from audiodag.signal.components.tonal import SineComponent
from audiodag.signal.components.noise import NoiseComponent
from audiodag.signal.components.component import CompoundComponent


def template_noisy_sine(duration: int = 1000,
                        fs: int = 5000,
                        freq: int = 12,
                        noise_mag: float = -80) -> CompoundComponent:
    """

    :param: fs: Sampling rate, Hz.
    :param duration: Duration in ms.
    :param freq: Frequency of sine wave, Hz.
    :param noise_mag: Noise mag in db (relative to sine mag of 1v)
    :return:
    """
    sin = SineComponent(freq=freq,
                        mag=1,
                        start=100,
                        fs=fs,
                        duration=duration)
    noise = NoiseComponent(fs=fs,
                           duration=duration,
                           mag=db_to_lin(ref=1,
                                     db_change=noise_mag))

    return CompoundComponent([sin, noise])


if __name__ == "__main__":

    ev = template_noisy_sine()
    ev.plot(show=True,
            channels=True)

    ev.plot_subplots(show=True)



from abc import ABC, abstractmethod
from typing import Union

import numpy as np


class Envelope(ABC):
    def __init__(self,
                 fs: Union[int, None] = None,
                 **kwargs):
        self.fs = fs

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __call__(self, y: np.ndarray) -> np.ndarray:
        return self.f(y)

    def __str__(self):
        return f"{self.__name__}"

    @abstractmethod
    def f(self, y: np.ndarray) -> np.ndarray:
        pass



"""
TODO: Replace visual checks with proper tests.
"""

import copy
import unittest
from functools import partial
from unittest.mock import MagicMock

import numpy as np
from matplotlib import pyplot as plt

from audiodag.signal.components.component import Component, CompoundComponent
from audiodag.signal.components.noise import NoiseComponent
from audiodag.signal.components.tonal import SineComponent
from audiodag.signal.envelopes.templates import CosRiseEnvelope


class TestEvent(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ev_1 = Component(fs=10000,
                             duration=20,
                             mag=1,
                             clip=np.inf)

    def test_length_as_expected(self):
        self.assertEqual(len(self.ev_1.y), 200)

    def test_y_no_envelope(self):
        ev = Component(fs=10000,
                       duration=20,
                       mag=1,
                       clip=np.inf)
        ev.envelope = lambda x: x

        self.assertListEqual(list(ev.y), list(np.ones(shape=(200,))))

    def test_plot(self):
        self.ev_1.plot()
        plt.show()


class TestNoiseEvent(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ev_1 = NoiseComponent(fs=10000,
                                  duration=20,
                                  mag=1,
                                  clip=np.inf,
                                  cache=True)

    def test_invalid_noise_dist_raises_error(self):
        self.assertRaises(ValueError, lambda: NoiseComponent(fs=10000,
                                                             duration=20,
                                                             mag=1,
                                                             clip=np.inf,
                                                             dist='invalid'))

    def test_dist_statistics(self):
        ev_norm = NoiseComponent(fs=10000,
                                 duration=20,
                                 mag=1,
                                 clip=np.inf,
                                 dist='normal')

        ev_uniform = NoiseComponent(fs=10000,
                                    duration=20,
                                    mag=1,
                                    clip=np.inf,
                                    dist='uniform')

        ev_norm.plot()
        ev_uniform.plot()
        plt.show()

        self.assertAlmostEqual(float(np.mean(ev_norm.y)), 0.0, -1)
        self.assertLess(float(np.mean(ev_norm.y)), float(np.mean(ev_uniform.y)), 0)
        self.assertLess(float(np.std(ev_uniform.y)), float(np.std(ev_norm.y)))

    def test_consistent_y_each_regen(self):
        y_1 = copy.deepcopy(self.ev_1.y)

        self.ev_1.clear()

        y_2 = copy.deepcopy(self.ev_1.y)

        self.assertTrue(np.all(y_1 == y_2))

    def test_no_seed_different_y(self):
        ev_1 = NoiseComponent(fs=10000,
                              duration=20,
                              mag=1,
                              clip=np.inf)

        ev_2 = NoiseComponent(fs=10000,
                              duration=20,
                              mag=1,
                              clip=np.inf)

        self.assertFalse(ev_1 == ev_2)
        self.assertFalse(np.all(ev_1.y == ev_2.y))

    def test_same_seed_same_y(self):
        ev_1 = NoiseComponent(fs=10000,
                              duration=20,
                              mag=1,
                              clip=np.inf,
                              seed=1)

        ev_2 = NoiseComponent(fs=10000,
                              duration=20,
                              mag=1,
                              clip=np.inf,
                              seed=1)

        self.assertTrue(ev_1 == ev_2)
        self.assertTrue(np.all(ev_1.y == ev_2.y))

    def test_diff_seed_diff_y(self):
        ev_1 = NoiseComponent(fs=10000,
                              duration=20,
                              mag=1,
                              clip=np.inf,
                              seed=1)

        ev_2 = NoiseComponent(fs=10000,
                              duration=20,
                              mag=1,
                              clip=np.inf,
                              seed=2)

        self.assertFalse(ev_1 == ev_2)
        self.assertFalse(np.all(ev_1.y == ev_2.y))

    def test_plot(self):
        self.ev_1.plot()
        plt.show()


class TestSineEvent(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.ev_1 = SineComponent(fs=10000,
                                 duration=20,
                                 mag=1,
                                 clip=np.inf)

    def test_plot(self):
        self.ev_1.plot()
        plt.show()


class TestCompoundEvent(unittest.TestCase):

    @staticmethod
    def _mock_compound_event(fs, start, duration_pts) -> MagicMock:
        return MagicMock(fs=fs,
                         start=start,
                         duration_pts=duration_pts,
                         x=np.linspace(start, duration_pts, duration_pts))

    def test_sine_noise_multiply(self):
        sine_event = SineComponent(fs=10000)
        noise_event = NoiseComponent(mag=0.3,
                                     fs=10000)

        compound_event = sine_event * noise_event

        self.assertAlmostEqual(float(np.mean(sine_event.y)), float(np.mean(compound_event.y)), 1)

    def test_construct_from_list_of_2(self):
        sine_event = SineComponent()
        compound_event = CompoundComponent(events=[sine_event, sine_event])

        self.assertAlmostEqual(float(np.mean(sine_event.y)), float(np.mean(compound_event.y)), 0)
        # This should be same as weighting is equal proportions by default
        self.assertAlmostEqual(float(np.std(compound_event.y)), float(np.std(sine_event.y)))

    def test_construct_from_list_of_2_set_weights_that_sum_to_one(self):
        sine_event = SineComponent(weight=0.5,
                                   seed=123)
        sine_event_2 = SineComponent(weight=0.5,
                                     seed=231)
        compound_event = CompoundComponent(events=[sine_event, sine_event_2])

        self.assertAlmostEqual(float(np.mean(sine_event.y)), float(np.mean(compound_event.y)), 0)

        self.assertAlmostEqual(compound_event.events[0].weight, 0.5)
        self.assertAlmostEqual(compound_event.events[1].weight, 0.5)

    def test_construct_from_list_of_2_set_weights_that_dont_sum_to_one(self):
        sine_event = SineComponent(weight=1.0,
                                   seed=123)
        sine_event_2 = SineComponent(weight=0.5,
                                     seed=231)

        self.assertAlmostEqual(sine_event.weight, 1.0, 3)
        self.assertAlmostEqual(sine_event_2.weight, 0.5, 3)

        compound_event = CompoundComponent(events=[sine_event, sine_event_2])

        self.assertAlmostEqual(float(np.mean(sine_event.y)), float(np.mean(compound_event.y)), 0)

        self.assertAlmostEqual(compound_event.events[0].weight, 0.667, 3)
        self.assertAlmostEqual(compound_event.events[1].weight, 0.333, 3)

        # Check weight of original objects is unchanged
        self.assertAlmostEqual(sine_event.weight, 1.0, 3)
        self.assertAlmostEqual(sine_event_2.weight, 0.5, 3)

    def test_construct_from_list_of_3(self):
        sine_event = SineComponent()
        noise_event = NoiseComponent()
        compound_event = CompoundComponent(events=[sine_event, sine_event, noise_event])

        self.assertAlmostEqual(float(np.mean(sine_event.y)), float(np.mean(compound_event.y)), 0)

    def test_construct_from_offset_list_of_3(self):
        sine_event_1 = SineComponent(start=200,
                                     duration=1000)
        sine_event_2 = SineComponent(start=300,
                                     duration=1000)
        sine_event_3 = SineComponent(start=500,
                                     duration=1000)

        compound_event = CompoundComponent(events=[sine_event_1, sine_event_2, sine_event_3])

        sine_event_1.plot()
        sine_event_2.plot()
        sine_event_3.plot()
        compound_event.plot(show=True)
        compound_event.plot_subplots(show=True)

    def test_construct_from_offset_list_of_3_adjusted_start(self):
        sine_event_1 = SineComponent(start=200,
                                     duration=1000)
        sine_event_2 = SineComponent(start=300,
                                     duration=1000)
        sine_event_3 = SineComponent(start=500,
                                     duration=1000)

        compound_event = CompoundComponent(events=[sine_event_1, sine_event_2, sine_event_3],
                                           start=0)

        sine_event_1.plot()
        sine_event_2.plot()
        sine_event_3.plot()
        compound_event.plot(show=True)
        compound_event.plot_subplots(show=True)

    def test_incompatible_events_fs_raises_error(self):
        sine_event = SineComponent()
        noise_event = NoiseComponent(fs=100)

        self.assertRaises(ValueError, lambda: CompoundComponent(events=[sine_event, sine_event, noise_event]))

    def test_create_from_fully_overlapping_long_list(self):
        rng = np.random.RandomState(123)

        ev_kwargs = {'fs': 2000,
                     'duration': 100,
                     'envelope': partial(CosRiseEnvelope,
                                         rise=10)}

        n = 40
        evs = [SineComponent(freq=e_i, **ev_kwargs) for e_i in rng.randint(5, 10, n)]

        compound_event = CompoundComponent(evs)
        compound_event.plot(channels=True,
                            show=True)

        self.assertEqual(compound_event.y.shape[0], 200)
        self.assertEqual(compound_event.channels().shape[0], n)
        self.assertEqual(compound_event.channels().shape[1], 200)

    def test_create_from_non_overlapping_long_list(self):
        rng = np.random.RandomState(123)

        ev_kwargs = {'fs': 2000,
                     'duration': 100}

        n = 6
        evs = [SineComponent(freq=f,
                             start=s_i * 100,
                             **ev_kwargs) for s_i, f in enumerate(rng.randint(5, 10, 6))]

        compound_event = CompoundComponent(evs)
        compound_event.plot(channels=True,
                            show=True)

        self.assertEqual(compound_event.y.shape[0], 200 * n)
        self.assertEqual(compound_event.channels().shape[0], n)
        self.assertEqual(compound_event.channels().shape[1], 200 * n)

    def test_create_from_partially_overlapping_long_list(self):
        ev_kwargs = {'fs': 2000,
                     'duration': 100}

        n = 3
        start_step = 80
        evs = [SineComponent(freq=10,
                             start=s_i * start_step,
                             **ev_kwargs) for s_i in range(n)]

        compound_event = CompoundComponent(evs)
        compound_event.plot(channels=True,
                            show=True)

        self.assertEqual(compound_event.y.shape[0], 200 + 200 + 120)
        self.assertEqual(compound_event.channels().shape[0], n)
        self.assertEqual(compound_event.channels().shape[1], 200 + 200 + 120)

    def test_compound_plus_normal_event(self):

        ev_kwargs = {'fs': 2000,
                     'duration': 100}

        n = 3
        start_step = 80
        evs = [SineComponent(freq=10,
                             start=s_i * start_step,
                             **ev_kwargs) for s_i in range(n)]

        compound_event_1 = CompoundComponent(evs)

        noise_event = NoiseComponent(mag=0.06,
                                     fs=2000,
                                     duration = 400)

        compound_event_2 = CompoundComponent([compound_event_1, noise_event])
        compound_event_2.plot(show=True,
                              channels=True)

        compound_event_2.channels()

    def test_compound_plus_compound_event(self):
        ev_kwargs = {'fs': 2000,
                     'duration': 100}

        n = 3
        start_step = 80
        evs = [SineComponent(freq=10,
                             start=s_i * start_step,
                             **ev_kwargs) for s_i in range(n)]

        compound_event_1 = CompoundComponent(evs)

        noise_events = [NoiseComponent(mag=0.1,
                                       start=100,
                                       **ev_kwargs),
                        NoiseComponent(mag=0.06,
                                       start=300,
                                       **ev_kwargs)]

        compound_event_2 = CompoundComponent(noise_events)

        compound_event_3 = CompoundComponent([compound_event_1, compound_event_2])
        compound_event_3.plot(show=True,
                              channels=True)

        compound_event_3.channels()



import numpy as np

from audiodag.signal.components.component import Component


class NoiseComponent(Component):
    """Class specifically for noisy components."""
    def __init__(self,
                 dist: str = 'normal',
                 *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.dist = dist

    @property
    def dist(self):
        return self._dist

    @dist.setter
    def dist(self, dist: str) -> None:
        if dist.lower() not in ['normal', 'uniform']:
            raise ValueError(f"Dist {dist} is not valid.")

        self._dist = dist

    def _generate_f(self) -> np.ndarray:
        """Sample noise from the RandomState."""
        return getattr(self.state, self.dist)(size=(self.duration_pts,)) * self.mag



from abc import ABC, abstractmethod
from typing import Any


class ReprID(ABC):
    @abstractmethod
    def __repr__(self) -> str:
        """__repr__ is used for id and eq, it should be redefined in children."""
        pass

    def __hash__(self) -> int:
        """Hash assumes important parameters are included in __repr__."""
        return hash(self.__repr__())

    def __eq__(self, other: Any) -> bool:
        return self.__hash__() == other.__hash__()

    def __lt__(self, other: Any) -> bool:
        """Ordering is undefined, but this is necessary for things like np.unique()."""
        return self.__hash__() < other.__hash__()



from dataclasses import dataclass
from audiodag.graph.node import Node
from typing import List, Tuple


@dataclass
class Graph:
    nodes: List[Node]


if __name__ == "__main__":

    Node(data='node_1',
         )


import numpy as np

from audiodag.signal.digital.conversion import db_to_lin
from audiodag.signal.envelopes.templates import IncreasingEnvelope
from audiodag.signal.components.component import CompoundComponent
from audiodag.signal.components.noise import NoiseComponent
from audiodag.signal.components.tonal import SineComponent


def template_complex():
    """
    Sine waves of increasing complexity in 3 steps:
    4 Hz -> 4 + (6 and 2) ->  4 + (6 and 2) + (8 and 10)

    Overlayed with increasing noise.
    :return:
    """
    start = 0
    sine_4 = SineComponent(start=start,
                           duration=1400,
                           freq=4)

    start = 200
    duration = 600
    sine_2_6 = CompoundComponent([SineComponent(start=start,
                                                duration=duration,
                                                freq=2),
                                  SineComponent(start=start,
                                                duration=duration,
                                                freq=6)])

    start = 600
    duration = 1000
    sine_2_12 = CompoundComponent([SineComponent(start=start,
                                                 duration=duration,
                                                 freq=2),
                                   SineComponent(start=start,
                                                 duration=duration,
                                                 freq=12)])
    sine_8_10 = CompoundComponent([SineComponent(start=start,
                                                 duration=duration,
                                                 freq=8),
                                   SineComponent(start=start,
                                                 duration=duration,
                                                 freq=10),
                                   sine_2_12])

    noise = NoiseComponent(start=0,
                           duration=1000,
                           envelope=IncreasingEnvelope,
                           mag=db_to_lin(ref=1,
                                     db_change=-120))

    return CompoundComponent([sine_4, sine_2_6, sine_8_10, noise])


if __name__ == "__main__":
    ev = template_complex()

    ev.to_list()

    # Plot final signal
    ev.plot(show=True)
    # Plot components of compound components
    ev.plot_subplots(show=True)
    ev.events[1].plot_subplots(show=True)
    ev.events[2].plot_subplots(show=True)
    ev.events[2].events[2].plot_subplots(show=True)

    ev.to_list()



import setuptools

from audiodag import MAJOR, MINOR, PATCH


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="AudioDAG",  # Replace with your own username
    version=f"{MAJOR}.{MINOR}.{PATCH}",
    author="Gareth Jones",
    author_email="",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/garethjns/AudioDAG",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"],
    python_requires='>=3.6',
    install_requires=['numpy', 'seaborn', 'pytest'])

# Lazy auto version increment
PATCH += 1
with open("audiodag/__init__.py", "r") as fh:
    spec = fh.readlines()

new_init = []
for l in spec:
    if l.startswith('PATCH'):
        l = F"PATCH = {PATCH}\n"
    new_init.append(l)

with open("audiodag/__init__.py", "w") as fh:
    fh.writelines(new_init)


import numpy as np

from audiodag.signal.components.component import Component


class SineComponent(Component):
    """Class specifically for tonal components."""
    def __init__(self,
                 freq: int = 2000,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.freq = freq

    def _generate_f(self) -> np.ndarray:
        """Generate vector for components"""
        return np.sin(np.linspace(0, 4 * np.pi * self.freq, self.duration_pts)) * 0.5 * self.mag



import numpy as np


def ms_to_pts(t_ms: int, fs: int) -> int:
    """Convert time in ms to time in pts."""
    return int(fs * t_ms / 1000)


def pts_to_ms(t_pts: int, fs: int) -> int:
    """Convert time in pts to time in ms (to nearest whole)"""
    return int(np.round(t_pts * 1000 / fs))


def lin_to_db(ref: float, ratio: float) -> float:
    """
    Convert linear ratio to dB change.

    :param ref: Reference value
    :param ratio: Linear ratio change.
    :return: dB change.
    """
    return 100 * np.log10((ref * ratio) / ref)


def db_to_lin(ref: float, db_change: float) -> float:
    """
    Convert dB change to linear ratio.

    :param ref: Reference value
    :param db_change: dB change.
    :return: Linear ratio change.
    """
    return 10 ** (db_change / 100) / ref ** 2
